import logging
import math
import pickle
import time
from collections import defaultdict

import boto3

OBJECT_SIZE = 400000
log = logging.getLogger(__name__)


class DynamoDBStateStore(object):
    def __init__(self, name, dynamodb_region) -> None:
        self.dynamodb = boto3.resource('dynamodb', region_name=dynamodb_region)
        self.client = boto3.client('dynamodb', region_name=dynamodb_region)
        self.name = name
        self.dynamodb_region = dynamodb_region
        self.table = self.dynamodb.Table(name)

    def build_key(self, type, iden) -> str:
        """
        It builds a unique partition key. The key could be objects with __str__ method.
        """
        return f"{type} {iden}"

    def restore(self, keys) -> dict:
        """
        Fetch all under the same parition key(keys).
        ret: <dict of key to states>
        """
        first_items = self._get_first_partitions(keys)
        remaining_items = self._get_remaining_partitions(first_items)
        vals = self._merge_items(first_items, remaining_items)
        return vals

    def alert(self, name: str, msg: str, error: str):
        import pysensu_yelp
        result_dict = {
            'name': name,
            'runbook': '',
            'status': 1,
            'output': '\n'.join(msg, error),
            'team': 'compute-infra',
            'tip': '',
            'page': None,
            'notification_email': 'mingqiz@yelp.com',
            'irc_channels': None,
            'slack_channels': None,
            'alert_after': '1m',
            'check_every': '10m',
            'realert_every': -1,
            'ttl': None
        }
        pysensu_yelp.send_event(**result_dict)

    def _get_items(self, keys: list) -> object:
        items = []
        for i in range(0, len(keys), 100):
            count = 0
            cand_keys = keys[i:min(len(keys), i + 100)]
            while True:
                resp = self.client.batch_get_item(
                    RequestItems={
                        self.name: {
                            'Keys': cand_keys,
                            'ConsistentRead': True
                        },
                    }
                )
                items.extend(resp['Responses'][self.name])
                if resp['UnprocessedKeys'].get(self.name) and count < 10:
                    cand_keys = resp['UnprocessedKeys'][self.name]['Keys']
                    count += 1
                elif count >= 10:
                    error = Exception('failed to retrieve items from dynamodb\n{}'.format(resp))
                    self.alert(
                        'tron_dynamodb_restore_failure',
                        'tron failed to restore for unknown reason with keys {}'.format(cand_keys),
                        str(error)
                    )
                    raise error
                else:
                    break
        return items

    def _get_first_partitions(self, keys: list):
        new_keys = [{'key': {'S': key}, 'index': {'N': '0'}} for key in keys]
        return self._get_items(new_keys)

    def _get_remaining_partitions(self, items: list):
        keys_for_remaining_items = []
        for item in items:
            remaining_items = [{'key': {'S': str(item['key']['S'])}, 'index': {'N': str(i)}}
                               for i in range(1, int(item['num_partitions']['N']))]
            keys_for_remaining_items.extend(remaining_items)
        return self._get_items(keys_for_remaining_items)

    def _merge_items(self, first_items, remaining_items) -> dict:
        items = defaultdict(list)
        raw_items = defaultdict(bytearray)
        # Merge all items based their keys and deserialize their values
        if remaining_items:
            first_items.extend(remaining_items)
        for item in first_items:
            key = item['key']['S']
            items[key].append(item)
        for key, item in items.items():
            item.sort(key=lambda x: int(x['index']['N']))
            for val in item:
                raw_items[key] += bytes(val['val']['B'])
        deserialized_items = {k: pickle.loads(v) for k, v in raw_items.items()}
        return deserialized_items

    def save(self, key_value_pairs) -> None:
        """
        Remove all previous data with the same partition key, examine the size of state_data,
        and splice it into different parts under 400KB with different sort keys,
        and save them under the same partition key built.
        """
        try:
            for key, val in key_value_pairs:
                self._delete_item(key)
                self[key] = pickle.dumps(val)
        except Exception as e:
            log.error(str(e))
            self.alert(
                'tron_dynamodb_save_failure',
                'tron failed to save for unknown reason with keys {}'.format(str(key_value_pairs.keys())),
                str(e)
            )

    def __setitem__(self, key: str, val: bytes) -> None:
        """
        Partition the item and write up to 10 partitions atomically.
        Retry up to 3 times on failure
        """
        num_partitions = math.ceil(len(val) / OBJECT_SIZE)
        items = []
        for index in range(num_partitions):
            item = {
                'Put': {
                    'Item': {
                        'key': {
                            'S': key,
                        },
                        'index': {
                            'N': str(index),
                        },
                        'val': {
                            'B': val[index * OBJECT_SIZE:min(index * OBJECT_SIZE + OBJECT_SIZE, len(val))],
                        },
                        'num_partitions': {
                            'N': str(num_partitions),
                        }
                    },
                    'TableName': self.name,
                },
            }
            count = 0
            resp = None
            items.append(item)
            # Only up to 10 items are allowed per transactions
            while len(items) == 10 or index == num_partitions - 1:
                try:
                    resp = self.client.transact_write_items(TransactItems=items)
                    items = []
                    break  # exit the while loop on successful writing
                except Exception as e:
                    count += 1
                    if count > 3:
                        log.error(str(e))
                        self.alert(
                            'tron_dynamodb_save_failure',
                            'tron failed to save due to transact_write_items failure with key {}\n{}'.format(key, resp),
                            str(e)
                        )
                        time.sleep(1)
                        break

    def _delete_item(self, key: str) -> None:
        try:
            with self.table.batch_writer() as batch:
                for index in range(self._get_num_of_partitions(key)):
                    batch.delete_item(
                        Key={
                            'key': key,
                            'index': index,
                        }
                    )
        except Exception as e:
            msg = 'Tron would work normally but redundant data might not be deleted if this is not resolved'
            log.error(str(e))
            self.alert(
                'tron_dynamodb_save_failure',
                'tron failed to delete for unknown reason {}\n{}'.format(key, msg),
                str(e)
            )

    def _get_num_of_partitions(self, key: str) -> int:
        """
        Return how many parts is the item partitioned into
        """
        try:
            partition = self.table.get_item(
                Key={
                    'key': key,
                    'index': 0,
                },
                ProjectionExpression='num_partitions',
                ConsistentRead=True
            )
            return int(partition.get('Item', {}).get('num_partitions', 0))
        except self.client.exceptions.ResourceNotFoundException:
            return 0

    def cleanup(self) -> None:
        return



#!/usr/bin/env python3.6
import argparse
import logging
import signal
import sys
from queue import Queue

import psutil
from twisted.internet import inotify
from twisted.internet import reactor
from twisted.python import filepath

from tron import yaml

log = logging.getLogger('tron.recover_batch')


class StatusFileWatcher(object):
    """
    Watches the status file produced by action runners
    """

    def __init__(self, to_watch, callback):
        notifier = inotify.INotify()
        notifier.startReading()
        notifier.watch(filepath.FilePath(to_watch), callbacks=[callback])


def parse_args():
    parser = argparse.ArgumentParser(
        description='Check if a action runner has exited; wait otherwise',
    )
    parser.add_argument('filepath', )
    return parser.parse_args()


def notify(notify_queue, ignored, filepath, mask):
    with open(filepath.path) as f:
        last_line = f.readlines()[-1]
        entries = yaml.load(last_line)

        pid = entries.get('runner_pid')
        return_code = entries.get('return_code')
        exit_code, error_msg = None, None

        if return_code is not None:
            if return_code < 0:
                # from the subprocess docs on the return code of a process:
                # "A negative value -N indicates that the child was terminated by signal N (POSIX only)."
                # We should always exit with a positive code, so we take the absolute value of the return code
                exit_code = abs(return_code)
                error_msg = (
                    'Action run killed by signal '
                    f'{signal.Signals(exit_code).name}'
                )
            else:
                exit_code = return_code

        elif not psutil.pid_exists(pid):
            exit_code = 1
            error_msg = (
                f'Action runner pid {pid} no longer running; '
                'unable to recover it'
            )

        if exit_code is not None:
            reactor.stop()
            notify_queue.put((exit_code, error_msg))


def get_key_from_last_line(filepath, key):
    with open(filepath) as f:
        lines = f.readlines()
        if lines:
            content = yaml.load(lines[-1])
            return content.get(key)
        return None


def run(fpath):
    existing_return_code = get_key_from_last_line(fpath, 'return_code')
    if existing_return_code is not None:
        sys.exit(existing_return_code)

    runner_pid = get_key_from_last_line(fpath, 'runner_pid')
    if not psutil.pid_exists(runner_pid):
        log.warning(
            f"Action runner pid {runner_pid} no longer running; "
            "unable to recover it"
        )
        #TODO: should we kill the process here?
        sys.exit(1)

    notify_queue = Queue()
    StatusFileWatcher(
        fpath,
        lambda *args, **kwargs: notify(notify_queue, *args, **kwargs)
    )
    reactor.run()
    exit_code, error_msg = notify_queue.get()
    if error_msg is not None:
        log.warning(error_msg)
    sys.exit(exit_code)


if __name__ == "__main__":
    args = parse_args()
    run(args.filepath)



from queue import Queue

import mock
import pytest

from tron.bin import recover_batch


@mock.patch('builtins.open', autospec=True)
@mock.patch.object(recover_batch, 'reactor', autospec=True)
@pytest.mark.parametrize('line,exit_code,error_msg', [
    (  # action runner finishes successfully
        '\n'.join(['return_code: 0', 'runner_pid: 12345']),
        0,
        None
    ), (  # action runner is killed
        '\n'.join(['return_code: -9', 'runner_pid: 12345']),
        9,
        'Action run killed by signal SIGKILL',
    ), (  # action runner is somehow no longer running
        'runner_pid: 12345',
        1,
        'Action runner pid 12345 no longer running; unable to recover it'
    )
])
def test_notify(mock_reactor, mock_open, line, exit_code, error_msg):
    fake_path = mock.MagicMock()
    mock_open.return_value.__enter__.return_value.readlines.return_value = [line]
    q = Queue()

    recover_batch.notify(q, 'some_ignored', fake_path, 'a_mask')

    actual_exit_code, actual_error_msg = q.get_nowait()
    assert actual_exit_code == exit_code
    assert actual_error_msg == error_msg
    assert mock_reactor.stop.call_count == 1


@mock.patch.object(recover_batch, 'get_key_from_last_line', mock.Mock(return_value=234))
def test_run_already_returned():
    with pytest.raises(SystemExit) as exc_info:
        recover_batch.run('a_path')

    assert exc_info.value.code == 234


@mock.patch.object(recover_batch, 'get_key_from_last_line', mock.Mock(side_effect=[
    None,  # return_code
    'a_pid',
]))
@mock.patch('psutil.pid_exists', mock.Mock(return_value=False), autospec=None)
def test_run_not_running():
    with pytest.raises(SystemExit) as exc_info:
        recover_batch.run('a_path')

    assert exc_info.value.code == 1


@mock.patch.object(recover_batch, 'get_key_from_last_line', mock.Mock(side_effect=[None, None]))
@mock.patch('psutil.pid_exists', mock.Mock(return_value=True), autospec=None)
@mock.patch.object(recover_batch, 'Queue', autospec=True)
@mock.patch.object(recover_batch, 'StatusFileWatcher', mock.Mock())
@mock.patch.object(recover_batch, 'reactor', mock.Mock())
def test_run_end_after_notify(mock_queue):
    mock_queue.return_value.get.return_value = (42, 'a_message')

    with pytest.raises(SystemExit) as exc_info:
        recover_batch.run('a_path')

    assert exc_info.value.code == 42



#!/usr/bin/env python
import argparse
import logging
import os
from datetime import datetime

import boto3

from tron.config import manager
from tron.config import schema

# Default values for arguments
DEFAULT_WORKING_DIR = '/var/lib/tron/'
DEFAULT_CONF_PATH = 'config/'
MAX_BACKUPS = 5
log = logging.getLogger('ddbbackup')


def create_backup(ddb, table_name):
    backup_name = '{}-{}'.format(table_name, datetime.now().strftime('%Y%m%d%H%M'))
    log.debug("Creating backup:", backup_name)
    response = ddb.create_backup(
        TableName=table_name, BackupName=backup_name)
    log.debug(response)


def delete_old_backups(ddb, table_name, max_backups):
    log.debug("Deleting old backups for table:", table_name)

    backups = ddb.list_backups(TableName=table_name)

    backup_count = len(backups['BackupSummaries'])
    log.debug('Total backup count:', backup_count)

    if backup_count <= max_backups:
        print("No stale backups. Exiting.")
        return

    sorted_list = sorted(backups['BackupSummaries'],
                         key=lambda k: k['BackupCreationDateTime'])

    old_backups = sorted_list[:max_backups]

    for backup in old_backups:
        arn = backup['BackupArn']
        log.debug("ARN to delete: " + arn)
        deleted_arn = ddb.delete_backup(BackupArn=arn)
        status = deleted_arn['BackupDescription']['BackupDetails']['BackupStatus']
        log.debug("Status:", status)

    return


def parse_cli():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-w",
        "--working-dir",
        default=DEFAULT_WORKING_DIR,
        help="Working directory for the Tron daemon, default %(default)s",
    )
    parser.add_argument(
        "-c",
        "--config-path",
        default=DEFAULT_CONF_PATH,
        help="File path to the Tron configuration file",
    )
    parser.add_argument(
        "-m",
        "--max-backups",
        default=MAX_BACKUPS,
        help="The maximum number of backups",
    )
    args = parser.parse_args()
    args.working_dir = os.path.abspath(args.working_dir)
    args.config_path = os.path.join(
        args.working_dir,
        args.config_path,
    )
    return args


def read_config(args):
    return manager.ConfigManager(args.config_path).load().get_master().state_persistence


def main():
    # Fetch configs. You can find the arguments in puppet.
    args = parse_cli()
    persistence_config = read_config(args)

    # If backup only if datastore is dynamodb
    store_type = schema.StatePersistenceTypes(persistence_config.store_type)
    if store_type == schema.StatePersistenceTypes.dynamodb:
        #set up dynamodb client
        region_name = persistence_config.dynamodb_region
        table_name = persistence_config.table_name
        max_backups = args.max_backups
        ddb = boto3.client('dynamodb', region_name=region_name)
        #create backup and remove old ones
        create_backup(ddb, table_name)
        delete_old_backups(ddb, table_name, max_backups)


if __name__ == '__main__':
    main()



from __future__ import absolute_import
from __future__ import unicode_literals

import logging

from tron.actioncommand import NoActionRunnerFactory
from tron.core.actionrun import ActionRun
from tron.core.actionrun import MesosActionRun
from tron.core.actionrun import SSHActionRun

log = logging.getLogger(__name__)


def filter_action_runs_needing_recovery(action_runs):
    ssh_runs = []
    mesos_runs = []
    for action_run in action_runs:
        if isinstance(action_run, SSHActionRun):
            if action_run.state == ActionRun.UNKNOWN:
                ssh_runs.append(action_run)
        elif isinstance(action_run, MesosActionRun):
            if action_run.state == ActionRun.UNKNOWN and action_run.end_time is None:
                mesos_runs.append(action_run)
    return ssh_runs, mesos_runs


def build_recovery_command(recovery_binary, path):
    return f"{recovery_binary} {path}"


def recover_action_run(action_run, action_runner):
    log.info(f"Creating recovery run for actionrun {action_run.id}")
    if type(action_runner) == NoActionRunnerFactory:
        log.info(
            f"Unable to recover action_run {action_run.id}: "
            "action_run has no action_runner"
        )
        return None

    recovery_run = SSHActionRun(
        job_run_id=action_run.job_run_id,
        name=f"recovery-{action_run.id}",
        node=action_run.node,
        bare_command=build_recovery_command(
            recovery_binary=f"{action_runner.exec_path}/recover_batch.py",
            path=f"{action_runner.status_path}/{action_run.id}/status",
        ),
        output_path=action_run.output_path,
    )
    recovery_action_command = recovery_run.build_action_command()
    recovery_action_command.write_stdout(
        f"Recovering action run {action_run.id}",
    )
    # Put action command in "running" state so if it fails to connect
    # and exits with no exit code, the real action run will not retry.
    recovery_action_command.started()

    # this line is where the magic happens.
    # the action run watches another actioncommand,
    # and updates its internal state according to its result.
    action_run.watch(recovery_action_command)

    if not action_run.machine.check('running'):
        log.error(
            f'Unable to transition action run {action_run.id} '
            f'from {action_run.machine.state} to start'
        )
    else:
        action_run.exit_status = None
        action_run.end_time = None
        action_run.machine.transition('running')

    log.info(
        f"Submitting recovery job with command {recovery_action_command.command} "
        f"to node {recovery_run.node}"
    )
    deferred = recovery_run.node.submit_command(recovery_action_command)
    deferred.addCallback(
        lambda x: log.info(f"Completed recovery run {recovery_run.id}")
    )
    return deferred


def launch_recovery_actionruns_for_job_runs(job_runs, master_action_runner):
    for run in job_runs:
        if not run._action_runs:
            log.info(f'Skipping recovery of {run} with no action runs (may have been cleaned up)')
            continue

        ssh_runs, mesos_runs = filter_action_runs_needing_recovery(run._action_runs)
        for action_run in ssh_runs:
            if type(action_run.action_runner) == NoActionRunnerFactory and \
               type(master_action_runner) != NoActionRunnerFactory:
                action_runner = master_action_runner
            else:
                action_runner = action_run.action_runner
            deferred = recover_action_run(action_run, action_runner)
            if not deferred:
                log.debug("unable to recover action run %s" % action_run.id)

        for action_run in mesos_runs:
            action_run.recover()
